# Copyright (c) 2014, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of exdesi.
#
# exdesi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exdesi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exdesi.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-
import os


def print_experiment_table(experimental_design) :

  nets         = set()
  readouts     = set()
  experiments  = set()
  classes      = 0
  pertubations = []
  difftables2  = dict()
  for a in experimental_design:
    if a.pred() == "counteqclasses" :
      classes = a.arg(0)
    if a.pred() == "pert" : # pert(E,G,S)
      pertubations.append((a.arg(0),a.arg(1),a.arg(2)))
    if a.pred() == "difflabel" :
      experiment = a.arg(0)
      net        = a.arg(1)
      readout    = a.arg(2)
      sign       = a.arg(3)
      experiments.add(experiment)
      nets.add(net)
      readouts.add((experiment,readout))
      #print('add:',experiment,net,readout,sign)
      difftables2[(experiment,net,readout)] = sign
  lo_nets        = sorted(list(nets))
  lo_readouts    = list(readouts)
  lo_experiments = list(experiments)
  lo_experiments.sort()
  print('We can differentiate',classes,'classes of networks with',len(experiments),'experiments.')

  for e in lo_experiments :
    print ('experiment',e,': ',end='')
    for p in pertubations:
      if p[0]==e : print (p[1],'=',p[2],end=', ')
    print ('')
  print ('')

  #print prediction table header
  for e in lo_experiments : print ('experiment',e,end=' | ')
  print ('')
  for e in lo_experiments:
    for r in lo_readouts :
      if r[0]==e : print (r[1],end=' ')
    print ('| ',end='')
  print('Network')

  #print prediction table content
  rows = []
  for n in lo_nets :
    row = []
    for e in lo_experiments:
      for r in lo_readouts :
        if r[0]==e : row.append(difftables2[(e,n,r[1])])
      row.append('|')
    row.append(n)
    rows.append(row)

  rows.sort()
  for r in rows :
    for e in r :
      if e == "1"   : print("  1 ",end='')
      elif e == '0' : print("  0 ",end='')
      elif e == "|" : print("|",end='')
      else: print('',e,'',end='')
    print('')


def clean_up() :
  if os.path.isfile("parser.out")              : os.remove("parser.out")
  if os.path.isfile("asp_py_lextab.py")        : os.remove("asp_py_lextab.py")
  if os.path.isfile("asp_py_lextab.pyc")       : os.remove("asp_py_lextab.pyc")
  if os.path.isfile("asp_py_parsetab.py")      : os.remove("asp_py_parsetab.py")
  if os.path.isfile("asp_py_parsetab.pyc")     : os.remove("asp_py_parsetab.pyc")
  if os.path.isfile("graph_parser_lextab.py")  : os.remove("graph_parser_lextab.py")
  if os.path.isfile("graph_parser_parsetab.py"): os.remove("graph_parser_parsetab.py")
  if os.path.isfile("parsetab.py")             : os.remove("parsetab.py")
  if os.path.isfile("parsetab.pyc")            : os.remove("parsetab.pyc")
  if os.path.isfile("sif_parser_lextab.py")    : os.remove("sif_parser_lextab.py")
  if os.path.isfile("sif_parser_lextab.pyc")   : os.remove("sif_parser_lextab.pyc")



#!python
# Copyright (c) 2014, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of exdesi.
#
# exdesi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exdesi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iggy.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-
import sys
import argparse
from pyasp.asp import *
from __exdesi__ import query, utils, bioquali

if __name__ == '__main__':

  parser = argparse.ArgumentParser()

  parser.add_argument("networkfiles",
    help="directory of influence graphs in SIF format")

  parser.add_argument("experivarfile",
    help="experimental variables")

  parser.add_argument('--best_set',type=int, default=-1,
    help="compute best set of experiments maximal number of experiments, default is OFF, 0=unlimited")

  parser.add_argument("-x", "--exclude",
    help="exclude experiments described in file EXCLUDE")

  args       = parser.parse_args()
  net_dir    = args.networkfiles
  exp_string = args.experivarfile
  flist      = os.listdir(net_dir)
  NETS       = TermSet()

  print('\nReading',len(flist),'networks from',net_dir,'...',end='\n')
  for f in flist :
    net_string = os.path.join(net_dir,f)
    print ('   reading',net_string,'... ',end='')
    net  = bioquali.readSIFGraph(net_string)
    NETS = TermSet(NETS.union(net))
    print('done.')

  print('\nReading experimental variables',exp_string, '... ',end='')
  mu = bioquali.readExpVar(exp_string)
  print('done.')
  #print(mu)

  if (args.exclude) :
    print('\nReading excluded experiments',args.exclude, '... ',end='')
    exclude = bioquali.readExcludedExp(args.exclude)
    print('done.')
    #print(exclude)
    MU = TermSet(mu.union(exclude))
  else : MU = mu

  print('\nCompute best single experiment ...',end='')
  experiments = query.get_best_single_experiments(NETS, MU)
  print('done.')

  if experiments == [] :
    print("no experiment can distinguish the networks.")
    print("add more readouts or more perturbations.")
  else:
    count = 0
    for e in experiments :
      count = count+1
      print("best single experiment",count,":")
      utils.print_experiment_table(e)

    if args.best_set > -1 :
      max_number_experiments = args.best_set
      print('\nCompute best experiment sets (max experiments = '+str(max_number_experiments)+') ...',end='')
      max_number_experiments = 10
      experiments = query.get_best_experiment_sets(NETS,MU,max_number_experiments)
      print('done.')

      count = 0
      for e in experiments :
        count = count+1
        print("best experiment set",count,":")
        utils.print_experiment_table(e)

  utils.clean_up()





# Copyright (c) 2014, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of exdesi.
#
# exdesi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exdesi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exdesi.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-
import re
from pyasp.asp import *
from pyasp.misc import *

import pyasp.ply.lex as lex
import pyasp.ply.yacc as yacc

import __exdesi__.sif_parser as sif_parser

def parse_val(s):
  if s == '+': return '1'
  elif s == '-': return '-1'
  elif s == '0': return '0'
  elif s == 'readout': return 'readout'
  else: 
    print(s)
    assert(False)


def readSIFGraph(filename):
  p = sif_parser.Parser(filename)
  """
  input: string, name of a file containing a Bioquali-like graph description
  output: asp.TermSet, with atoms matching the contents of the input file
  
  Parses a Bioquali-like graph description, and returns
  a TermSet object.
  Written using original Bioquali
  """

  accu = TermSet()
  file = open(filename,'r')
  s = file.readline()
  while s!="":
    try: accu = p.parse(s)
    except EOFError: break
    s = file.readline()
  return accu


#read experiment variables (possible perturbations, readouts)
def readExpVar(filename):
  GENE_ID     = '[-a-zA-Z0-9_:\(\)/]+'
  VAL         = '(-|\+|0|nc|readout|input|notPlus|notMinus)'
  file        = open(filename,'r')
  val_re      = '(?P<genid>'+GENE_ID+')(\s)*=(\s)*(?P<sign>'+VAL+')'
  val         = re.compile(val_re)
  line_number = 1
  line        = file.readline()
  accu        = TermSet()

  while line:
    if line.strip() :
      vm = val.match(line)
      if vm:
        if parse_val(vm.group('sign'))=='readout':
          vertex = quote(vm.group('genid'))
          accu.add(Term('preadout',["gen("+vertex+")"]))
        else:
          vertex = quote(vm.group('genid'))
          accu.add(Term('pperturb',["gen("+vertex+")", parse_val(vm.group('sign'))]))
      else:
        print('Syntax error line:', line_number, ':', line)

      line = file.readline()
      line_number+=1
    else:    
      line = file.readline()
      line_number+=1        
  return accu

#read excluded experiments
def readExcludedExp(filename):

  GENE_ID     = '[-a-zA-Z0-9_:\(\)/]+'
  VAL         = '(-|\+|0|nc|readout|input|notPlus|notMinus)'
  file        = open(filename,'r')
  val_re      = '(?P<genid>'+GENE_ID+')(\s)*=(\s)*(?P<sign>'+VAL+')'
  val         = re.compile(val_re)
  line_number = 1
  line        = file.readline()
  accu        = TermSet()
  token       = 'experiment'
  expident_re = 'experiment(\s)+(?P<expid>'+GENE_ID+'):'
  expident    = re.compile(expident_re)
  expid       = '1'

  while line:
     if line.strip() :
       vm = expident.match(line)
       if vm : 
         expid = vm.group('expid')
         line  = file.readline()
         line_number+=1
       else:
         vm = val.match(line)
         if vm:
             if parse_val(vm.group('sign'))=='readout':
               print('Warning line:',line_number,': Readouts are currently not handled, in experiment exclusion!')
             else:
               vertex = quote(vm.group('genid'))
               accu.add(Term('done',[expid,"gen("+vertex+")", parse_val(vm.group('sign'))]))
         else:
           print('Syntax error line:', line_number, ':', line)

         line = file.readline()
         line_number+=1
     else:    
       line = file.readline()
       line_number+=1
  return accu





# Copyright (c) 2014, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of exdesi.
#
# exdesi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exdesi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exdesi.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

"""
This module contains the queries which can be asked to the model and data.

"""
import os
import tempfile
from pyasp.term import *
from pyasp.asp import *

root = __file__.rsplit('/', 1)[0]
find_best_exp_sets_prg = root + '/encodings/find_best_experiment_sets.lp'
elem_path_prg          = root + '/encodings/elementary_path_constraint.lp'


def get_best_single_experiments(nets,expvars):
  '''
  returns the experiments as a``TermSet`` object [instance].
  '''
  netsf    = nets.to_file()
  expvarsf = expvars.to_file()

  i = 1 #single experiment

  num_exp   = String2TermSet('pexperiment('+str(i)+')')
  num_expf  = num_exp.to_file()
  prg       = [ netsf, expvarsf, num_expf, find_best_exp_sets_prg ,
                elem_path_prg ]
  coptions  = '--project --opt-mode=optN --opt-strategy=0 --opt-heuristic'
  solver    = GringoClasp(clasp_options=coptions)
  solutions = solver.run(prg,collapseTerms=True,collapseAtoms=False)

  os.unlink(num_expf)

  os.unlink(netsf)
  os.unlink(expvarsf)
  return solutions


def get_best_experiment_sets(nets,expvars,num):
  '''
  given the network and the experimental variables, and the bound on the size of an experiment set
  returns the experiments as a``TermSet`` object [instance].
  '''

  netsf    = nets.to_file()
  expvarsf = expvars.to_file()

  best           = -1
  best_solutions = []
  best_found     = False

  i = 0
  while i < num and not best_found :
    i += 1

    num_exp   = String2TermSet('pexperiment('+str(i)+')')
    num_expf  = num_exp.to_file()
    prg       = [ netsf, expvarsf, num_expf, find_best_exp_sets_prg,
                  elem_path_prg ]
    coptions  = '--project --opt-mode=optN --opt-strategy=0 --opt-heuristic'
    solver    = GringoClasp(clasp_options=coptions)
    solutions = solver.run(prg,collapseTerms=True,collapseAtoms=False)

    #print(solutions[0].score[0],solutions[0].score[1],solutions[0].score[2],solutions[0].score[3])

    os.unlink(num_expf)
    if solutions == []: best_found = True
    else:
      opt=(solutions[0].score[0]+solutions[0].score[1]+solutions[0].score[2])
      if best == opt:
        best_found = True
      else:
        best           = opt
        best_solutions = solutions


  os.unlink(netsf)
  os.unlink(expvarsf)

  return best_solutions





# Copyright (c) 2014, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of exdesi.
#
# exdesi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exdesi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exdesi.  If not, see <http://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

from pyasp.asp import *
from pyasp.misc import *

class Lexer:
  tokens = (
    'IDENT',
    'NOT',
    'AND',
    'LB',
    'RB',
    'PLUS',
    'MINUS',
  )
  
  # Tokens
  
  t_IDENT = r'[a-zA-Z][a-zA-Z0-9_:\-\[\]/]*'
  t_NOT   = r'!'
  t_AND   = r'&'
  t_LB    = r'\('
  t_RB    = r'\)'
  t_PLUS  = r'1'
  t_MINUS = r'-1'


  def __init__(self):
    import pyasp.ply.lex as lex
    self.lexer = lex.lex(object = self, optimize=1, lextab='sif_parser_lextab')

  # Ignored characters
  t_ignore = " \t"

  def t_newline(self, t):
    r'\n+'
    t.lexer.lineno += t.value.count("\n")
	
  def t_error(self, t):
    print("Illegal character '",str(t.value[0]),"'", sep='')
    t.lexer.skip(1)


class Parser:
  tokens     = Lexer.tokens
  precedence = ( )

  def __init__(self, name):
    self.name   = quote(name)
    self.accu   = TermSet()
    self.args   = []
    self.lexer  = Lexer()
    import pyasp.ply.yacc as yacc
    self.parser = yacc.yacc(module=self, tabmodule='calc_parsetab', debugfile="calc_parser.out")
    #self.parser = yacc.yacc(module=self,optimize=1,debug=0, write_tables=0)

  def p_statement_expr(self, t):
    '''statement : node_expression PLUS node_expression
                 | node_expression MINUS node_expression
                 | node_expression PLUS andnode_expression
                 | node_expression MINUS andnode_expression
                 | andnode_expression PLUS node_expression
                 | andnode_expression MINUS node_expression
                 | andnode_expression PLUS andnode_expression
                 | andnode_expression MINUS andnode_expression
                 '''
    if len(t)<3 :
      self.accu.add(Term('input', [t[1]]))
      print('input', t[1])
    else:
      #print t[1], t[2], t[3]
      self.accu.add(Term('edge', [self.name,t[1],t[3]]))
      self.accu.add(Term('obs_elabel', [self.name,t[1],t[3],t[2]]))
      #print Term('obs_elabel', ["gen(\""+t[1]+"\")","gen(\""+t[3]+"\")",t[2]])


  def p_andnode_expression(self, t):
    '''andnode_expression : LB  identlist RB '''
    self.accu.add(Term('vertex', [self.name,"and(\""+t[2]+"\")"]))
    t[0] = "and(\""+t[2]+"\")"


  def p_identlist(self, t):
    '''identlist : IDENT
                  | NOT IDENT
                  | IDENT AND identlist
                  | NOT IDENT AND identlist
                  '''
    if len(t)==5 :
      #print(t[1],t[2],t[3],t[4])
      t[0] = t[1]+t[2]+t[3]+t[4]
    elif len(t)==4 :
      #print(t[1],t[2],t[3])
      t[0] = t[1]+t[2]+t[3]
    elif len(t)==3 :
      #print(t[1],t[2])
      t[0] = t[1]+t[2]
    elif len(t)==2 :
      #print(t[0],t[1])
      t[0]=t[1]
    else:
      print("Syntax error at '",str(t),"'")


  def p_node_expression(self, t):
    '''node_expression : IDENT'''
    self.accu.add(Term('vertex', [self.name,"gen(\""+t[1]+"\")"]))
    t[0] = "gen(\""+t[1]+"\")"


  def p_error(self, t):
    print("Syntax error at '",str(t),"'")

  def parse(self, line):
    self.parser.parse(line, lexer=self.lexer.lexer)
    return self.accu



# Copyright (c) 2014, Sven Thiele <sthiele78@gmail.com>
#
# This file is part of exdesi.
#
# exdesi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exdesi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exdesi.  If not, see <http://www.gnu.org/licenses/>.
# -*- coding: utf-8 -*-

from setuptools import setup
                         
setup(
  name             = 'exdesi',
  version          = '0.3',
  url              = 'http://bioasp.github.io/exdesi/',
  license          = 'GPLv3+',
  description      = 'Design experiments to differentiate interaction graph '
                     'models.',
  long_description = open('README.rst').read(),
  author           = 'Sven Thiele',
  author_email     = 'sthiele78@gmail.com',
  packages         = ['__exdesi__'],
  package_dir      = {'__exdesi__' : 'src'},
  package_data     = {'__exdesi__' : ['encodings/*.lp','encodings/*.gringo']},
  scripts          = ['exdesi.py'],
  install_requires = ['pyasp == 1.4.3']
)
